import unittest
import pandas as pd
import numpy as np
import weedout.preprocess as weedout

class TestInitialCheckDt(unittest.TestCase):
    def test_wrong_path(self):
        file_path = 'non_existent_file.csv'
        target_name = 'target_name'
        unscale_columns = []
        
        with self.assertRaises(Exception) as context:
            weedout.initial_check_dt(file_path, target_name, unscale_columns)
        
        self.assertEqual(str(context.exception), f"An error occurred: [Errno 2] No such file or directory: '{file_path}'") 
    
    def test_wrong_targetname(self):
        file_path = 'tests_static/regularCSVFile.csv'
        target_name = 'non_existent'
        unscale_columns = []

        with self.assertRaises(Exception) as context:
            weedout.initial_check_dt(file_path, target_name, unscale_columns)

        self.assertEqual(str(context.exception), f"The target variable '{target_name}' is missing in the DataFrame.") 
    
    def test_duplicate_columns(self):
        file_path = 'tests_static/duplicateCSVFile.csv'
        target_name = 'target'
        unscale_columns = []

        with self.assertRaises(Exception) as context:
            weedout.initial_check_dt(file_path, target_name, unscale_columns)

        self.assertEqual(str(context.exception), "An error occurred: ('Duplicate columns exist in the file:', ['feature2'])") 

    def test_column_drop(self):
        file_path = 'tests_static/regularCSVFile.csv'
        target_name = 'target'
        unscale_columns = ['feature1']

        df = weedout.initial_check_dt(file_path, target_name, unscale_columns)

        self.assertEqual(['id','feature2','feature3','target'], df.columns.to_list())

class TestCrossSectionalImputation(unittest.TestCase):

    def test_row_impute(self):
        pre_df = pd.read_csv('tests_static/csImputeCSVFile.csv')
        target_name = 'target'

        df = weedout.cross_sectional_imputation(pre_df, target_name)

        self.assertEqual([3.0, 5.6, 'sad', 0.0], df.iloc[2].tolist())


class TestTimeSeriesImputation(unittest.TestCase):

    def test_row_impute(self):
        pre_df = pd.read_csv('tests_static/tsImputeCSVFile.csv')

        target_name = 'target'

        df = weedout.time_series_imputation(pre_df, target_name)

        self.assertEqual([4.0,6.7,'sad',1.0], df.iloc[4].tolist())
    
    def test_missingnum_firstrow(self):
        target_name = 'target'

        with self.assertRaises(Exception) as context:
            pre_df = pd.read_csv('tests_static/tsImputeMissingCSVFile.csv')
            weedout.time_series_imputation(pre_df, target_name)

        self.assertEqual(str(context.exception), "The first row cannot have a numeric null value.") 

class TestHandleBalanceDataset(unittest.TestCase):

    def test_oversampling(self):

        pre_df = pd.read_csv('tests_static/balanceCSVFile.csv')
        target_variable='target'
        strategy = 'oversampling'

        def min_and_max_ratio(df, target):
            max_count = df[target].value_counts().max()
            min_count = df[target].value_counts().min()
            return min_count/max_count
        
        ratio_minmax_pre = min_and_max_ratio(pre_df, target_variable)
        df = weedout.handle_imbalanced_data(pre_df, target_variable, strategy)
        ratio_minmax_post = min_and_max_ratio(df, target_variable)
        self.assertGreater(ratio_minmax_post, ratio_minmax_pre)
    
    def test_undersampling(self):

        pre_df = pd.read_csv('tests_static/balanceCSVFile.csv')
        target_variable='target'
        strategy = 'undersampling'

        def min_and_max_ratio(df, target):
            max_count = df[target].value_counts().max()
            min_count = df[target].value_counts().min()
            return min_count/max_count
        
        ratio_minmax_pre = min_and_max_ratio(pre_df, target_variable)
        df = weedout.handle_imbalanced_data(pre_df, target_variable, strategy)
        ratio_minmax_post = min_and_max_ratio(df, target_variable)
        self.assertGreater(ratio_minmax_post, ratio_minmax_pre)
    
    def test_SMOTE(self):

        pre_df = pd.read_csv('tests_static/balanceCSVFile.csv')
        target_variable='target'
        strategy = 'smote'

        def min_and_max_ratio(df, target):
            max_count = df[target].value_counts().max()
            min_count = df[target].value_counts().min()
            return min_count/max_count
        
        ratio_minmax_pre = min_and_max_ratio(pre_df, target_variable)
        df = weedout.handle_imbalanced_data(pre_df, target_variable, strategy)
        ratio_minmax_post = min_and_max_ratio(df, target_variable)
        self.assertGreater(ratio_minmax_post, ratio_minmax_pre)


class TestRemoveOutliers(unittest.TestCase):

    def test_remove_outliers(self):
         # Load the data
        df1 = pd.read_csv('tests_static/outlierCSVFile.csv')
        df2 = weedout.remove_outliers(df1)

        # Merge DataFrames with indicator to find differing rows
        diff_df = df1.merge(df2, how='outer', indicator=True)
        rows_in_df1_not_in_df2 = diff_df[diff_df['_merge'] == 'left_only']

        # Define the rows to check
        rows_to_check = pd.DataFrame({
            'Name': ['Frank', 'Jack'],
            'Age': [28, 31],
            'Height': [400, 178],
            'Weight': [200, 900]
        })

        # Select relevant columns for comparison
        columns_to_compare = ['Name', 'Age', 'Height', 'Weight']
        rows_in_df1_not_in_df2 = rows_in_df1_not_in_df2[columns_to_compare]
        rows_to_check = rows_to_check[columns_to_compare]

        # Reset index for proper comparison
        rows_in_df1_not_in_df2 = rows_in_df1_not_in_df2.reset_index(drop=True)
        rows_to_check = rows_to_check.reset_index(drop=True)

        # Assert that the DataFrames are equal
        pd.testing.assert_frame_equal(rows_in_df1_not_in_df2, rows_to_check)

class TestSeperateTargetColumn(unittest.TestCase):

    def test_file_type(self):
        df='tests_static/duplicateCSVFile.csv'
        target_name = 'target'

        with self.assertRaises(Exception) as context:
            weedout.separate_target_column(df, target_name)

        self.assertEqual(str(context.exception), "Data type error.Please provide the right type of dataframe.") 

    def test_wrong_targetname(self):
        df=pd.read_csv('tests_static/duplicateCSVFile.csv')
        target_name = 'non_existent'

        with self.assertRaises(Exception) as context:
            weedout.separate_target_column(df,target_name)

        self.assertEqual(str(context.exception), f"Target Column does not Exist. Please provide the right one.")

    def test_regular_split(self):
        pre_df = pd.read_csv('tests_static/regularCSVFile.csv')
        target_name = 'target'
        df, target= weedout.separate_target_column(pre_df, target_name)
        self.assertEqual(['id','feature1','feature2','feature3'], df.columns.to_list())
        self.assertEqual(['target'], target.columns.tolist() )

class TestEncoding(unittest.TestCase):

    def test_encoding_no_object(self):
        pre_df = pd.read_csv('tests_static/regularCSVFile.csv')
        df = weedout.encoding(pre_df)
        pd.testing.assert_frame_equal(pre_df,df)

    def test_onehot_encoding(self):
        pre_df = pd.read_csv('tests_static/oneHEncodeCSVFile.csv')
        df = weedout.encoding(pre_df,['feature1'])
        self.assertEqual(5, len(df.columns.to_list()))
        self.assertEqual(['feature1', 'feature2', 'target','feature3_happy', 'feature3_sad'], df.columns.to_list())
    
    def test_label_encoding(self):
        pre_df = pd.read_csv('tests_static/labelEncodeCSVFile.csv')
        df = weedout.encoding(pre_df,['feature1'])
        self.assertEqual(4, len(df.columns.to_list()))
        self.assertEqual(['feature1', 'feature2', 'feature3','target'], df.columns.to_list())
        self.assertEqual('int', df['feature3'].dtype)
        self.assertEqual(4, df['feature3'].nunique())


class TestCombine(unittest.TestCase):

    def test_combine(self):
        pre_df = pd.read_csv('tests_static/regularCSVFile.csv')
        features, target = weedout.separate_target_column(pre_df, 'target')
        df = weedout.combine(features, target)
        pd.testing.assert_frame_equal(pre_df, df)

    def test_combine_fail(self):
        pre_df = pd.read_csv('tests_static/regularCSVFile.csv')
        with self.assertRaises(Exception) as context:
            pre_df = pd.read_csv('tests_static/tsImputeMissingCSVFile.csv')
            weedout.combine(pre_df, "target")
        self.assertEqual(str(context.exception), f"Data type error.Please provide the right type of dataframe.")

class TestFeatureScaling(unittest.TestCase):
        
        def test_featurescaling(self):
        
            n = 200

            normal_1 = np.random.normal(size=n)
            normal_2 = np.random.normal(size=n)

            exponential = np.random.exponential(size=n)
            lognormal = np.random.lognormal(size=n)
            binary_values = np.random.randint(2, size=n)

            pre_df = pd.DataFrame({
                'Normal_1': normal_1,
                'Exponential': exponential,
                'Lognormal': lognormal,
                'Normal_2': normal_2,
                'Binary': binary_values,
            })

            df = weedout.feature_scaling(pre_df, [])

            # MinMaxScaled Columns should result in values between 0 and 1. 
            self.assertEqual(0, df['Exponential'].min())
            self.assertAlmostEqual(1, df['Exponential'].max())
            self.assertEqual(0, df['Lognormal'].min())
            self.assertAlmostEqual(1, df['Lognormal'].max())

            # Binary values should not be worked upon.
            pd.testing.assert_frame_equal(df[['Binary']], pre_df[['Binary']])


class TestPipeline(unittest.TestCase):

    def test_final_pipeline(self):

        pre_df = pd.read_csv('tests_static/regularCSVFile.csv')
        target_variable ='target'
        strategy = "smote"
        sampling = 1
        classification = 1


        df = weedout.preprocess_pipeline('tests_static/regularCSVFile.csv', target_variable, ['feature1'], ['id'], type_dataset=0, sampling=1, classfication=1, strategy_sample=strategy)
        
        self.assertEqual('object', df['id'].dtype)
        self.assertTrue(not df.isnull().values.any())

        # Test for Smote
        if classification:
            def min_and_max_ratio(df, target):
                max_count = df[target].value_counts().max()
                min_count = df[target].value_counts().min()
                return min_count/max_count
            
            ratio_minmax_pre = min_and_max_ratio(pre_df, target_variable)
            ratio_minmax_post = min_and_max_ratio(df, target_variable)
            self.assertGreater(ratio_minmax_post, ratio_minmax_pre)

        # Test for no null values:
        if sampling:
            if classification:
                self.assertEqual(0, df.isnull().sum().sum())
        

if __name__ == '__main__':
    unittest.main()



from .preprocess import (
    check_duplicate_columns,
    initial_check_dt,
    cross_sectional_imputation,
    time_series_imputation,
    handle_imbalanced_data,
    remove_outliers,
    separate_target_column,
    encoding,
    feature_scaling,
    combine,
    preprocess_pipeline
)

from .visualization import (
    plot_filtered_correlation_matrix,
    show_removed_outliers,
)

from .datainfo import (
    filtered_correlation_matrix,
    display,
)

__all__ = [
    'check_duplicate_columns',
    'initial_check_dt',
    'cross_sectional_imputation',
    'time_series_imputation',
    'handle_imbalanced_data',
    'remove_outliers',
    'separate_target_column',
    'filtered_correlation_matrix',
    'plot_filtered_correlation_matrix',
    'encoding',
    'feature_scaling',
    'combine',
    'display',
    'preprocess_pipeline',
    'show_removed_outliers'
]


import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from statsmodels.stats.outliers_influence import variance_inflation_factor


def show_removed_outliers(df_original: pd.DataFrame, df_pre_processed: pd.DataFrame):
    """
        This is a visualization function to compare the boxplot spread of the original dataset and 
        the proccessed dataset. 

        @paramters:

            df_original: pd.DataFrame
                The dataframe for the original dataset.

            df_preprocessed: pd.DataFrame
                The dataframe for the preprocessed dataset.

    """

    sns.set(style="whitegrid")

    # Get numerical columns
    numerical_cols = df_original.select_dtypes(include=['int64', 'float64']).columns

    # Define the figure and axes for the subplots
    fig, axes = plt.subplots(nrows=2, ncols=len(numerical_cols), figsize=(50, 50), sharey=True)
    
    # Flatten the axes array for easy iteration
    axes = axes.flatten()

    # Plot box-plots for original and pre-processed dataframe
    for i, column in enumerate(numerical_cols):
        if i < len(axes) // 2:
            sns.boxplot(data=df_original[column], ax=axes[i], palette="Set2")
            axes[i].set_title(f'Original - {column}')
        else:
            sns.boxplot(data=df_pre_processed[column], ax=axes[i], palette="Set2")
            axes[i].set_title(f'Pre-Processed - {column}')

        axes[i].set_xlabel('')
        axes[i].set_ylabel('')

    # Adjust layout
    plt.tight_layout()
    plt.show()


def plot_filtered_correlation_matrix(df: pd.DataFrame):
    """
        This function plots the correlation matrix of the features based onVariance Inflation Factor ( VIF = 1/(1-R^2) ) 
        of each feature column. VIF is a strong indicator of multi-collinearity in our dataframe. 

        Note : The user is expected to remove some of the corelated features based on the VIF value.

        @paramater:

            df : pd.DataFrame
                The dataframe provided by user.
    
    """
    numerical_columns = df.select_dtypes(include=['float64', 'int64'])
    non_numerical_columns = df.select_dtypes(exclude=['float64', 'int64']).columns
    print("Before: ", numerical_columns.shape)
    
    vif = pd.DataFrame()
    vif['Feature'] = numerical_columns.columns
    vif["VIF"] = [variance_inflation_factor(numerical_columns.values, i) for i in range(numerical_columns.shape[1])]
    
    print("Initial VIF values: ")
    print(vif.sort_values(by="VIF", ascending=False))
    
    infinite_vif_features = vif[vif["VIF"] == np.inf]["Feature"].tolist()
    if infinite_vif_features:
        print(f"\nDropping columns with infinite VIF values: {infinite_vif_features}\n")
        numerical_columns = numerical_columns.drop(columns=infinite_vif_features)
        
    max_vif = 10
    remove_flag = True
    
    while remove_flag:
        vif = pd.DataFrame()
        vif['Feature'] = numerical_columns.columns
        vif["VIF"] = [variance_inflation_factor(numerical_columns.values, i) for i in range(numerical_columns.shape[1])]
        
        max_vif_feature = vif.loc[vif['VIF'].idxmax()]
        
        if max_vif_feature['VIF'] > max_vif:
            numerical_columns = numerical_columns.drop(max_vif_feature['Feature'], axis=1)
            print(f"Removed variable with high VIF {max_vif_feature['Feature']} (VIF={max_vif_feature['VIF']})")
        else:
            remove_flag = False

    print("After: ", numerical_columns.shape)
    
    plt.figure(figsize=(13,10))
    plt.title("VIF")
    sns.heatmap(numerical_columns.corr(),annot=True,fmt='0.2g',cmap='coolwarm',vmin=-1,vmax=1)
    plt.show()


import pandas as pd
import numpy as np
import csv
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import StandardScaler, MinMaxScaler,OneHotEncoder, FunctionTransformer
from statsmodels.stats.outliers_influence import variance_inflation_factor
from imblearn.over_sampling import RandomOverSampler, SMOTE
from imblearn.under_sampling import RandomUnderSampler
from sklearn.impute import SimpleImputer
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from scipy import stats
from typing import List, Optional, Tuple
from sklearn.utils import resample



def check_target_variable(df: pd.DataFrame, target_name: str) -> bool:
    """
        The function returns true if the given target name exists in the dataframe. It returns false
        if the target name is not in the dataframe. 

        @paramters: 
            df: pd.DataFrame
                Given dataframe.
            
            target_name: str
                Given target Name 
        @return:

            bool:
                Indicates if target name exists of not.
    """
    seen = False 
    for col in df.columns:
        if col == target_name:
            seen = True 
            break 
    return seen


def check_duplicate_columns(file_path: str) -> List[str]:
    """
        The function checks if the dataset has multiple same column names.

        @paramter:

            file_path:
                The path to the csv file 

        @return:

            List[str]:
                A list of all the duplicate column names.

    """
    if file_path.lower().endswith('.csv'):
        with open(file_path, 'r') as file:
            reader = csv.reader(file)
            headers = next(reader)
            seen = set()
            duplicates = set()
            
            for header in headers:
                if header in seen:
                    duplicates.add(header)
                seen.add(header)
            
            if duplicates:
                return list(duplicates)
            else:
                return []
    else:
        raise Exception('Wrong File Type. Only CSV allowed.')


def initial_check_dt(file_path: str, target_variable: str, columns_to_drop: List[str]) -> Optional[pd.DataFrame]:
    """
        The function initially checks if the file provided is a valid CSV file that can be parsed through. It checks 
        and drops duplicate columns, verifies the target variable, and drops columns with exceedingly high
        null values and untouched columns.

        @parameters:

            file_path : str
                The path to the csv file.

            target_variable : str  
                The target column of your dataset.

            untouched_columns : List[str] 
                The list of columns that are not going to be modified. These 
                would generally be columns that have to do with IDw.
        
        @return:

            Optional[pd.DataFrame] 
                It will return the modified dataframe if the function is successful.

    """
    if file_path.lower().endswith('.csv'):

        try:
            duplicate = check_duplicate_columns(file_path)
            if duplicate:
                raise Exception("Duplicate columns exist in the file:", duplicate)
            df = pd.read_csv(file_path)
        except pd.errors.ParserError:
            raise Exception('File cannot be parsed as a CSV.') from None
        except Exception as e:
            raise Exception(f"An error occurred: {e}") from e
        
        id_drop = [col for col in df.columns if col in columns_to_drop]
        if id_drop:
            df = df.drop(columns=id_drop)
            print(f"Dropped untouched columns: {', '.join(id_drop)}")
            
        drop_columns = []
        for col in df.columns:
            missing_count = df[col].isnull().sum()
            total_values = len(df[col])
            missing_percentage = (missing_count / total_values) * 100

            if missing_percentage > 40:
                drop_columns.append(col)

        print("Dropped columns with high missing data: ", list(drop_columns))
        df = df.drop(columns=drop_columns)
        
        df_columns = df.columns.tolist() 
        keep_indices = []

        for index, col in enumerate(df.columns):
            keep_indices.append(index)

        df = df.iloc[:, keep_indices]
        
        if target_variable in df_columns:
            print(f"The target variable '{target_variable}' is present in the DataFrame.")
        else:
            raise Exception (f"The target variable '{target_variable}' is missing in the DataFrame.")
        
        print("Initial checks and changes made")
        return df
    
    else:
        raise Exception("Oops :( The file is not a CSV file!")

def cross_sectional_imputation(cross_sectional_df: pd.DataFrame, target_name: str) -> pd.DataFrame:
    """
        The function works on imputation for the given cross sectional dataframe. For columns with
        numeric values, the null values of that column are imputed/filled with the mean (average) value 
        of the column. For columns with string values, the null values of that column are imputed/filled
        with the mode (most frequent) value of that column. 

        @parameters:

            cross_sectional_df : pd.DataFrame:
                The dataframe of the cross-sectional dataset with the target column included.

            target_name: str:
                The name of the target column
        
        @return:
        
            Optional[pd.DataFrame] 
                It will return the imputed dataframe if the function is successful.

    """
    df = cross_sectional_df.copy()
    
    print("Total Null value counts before imputation: \n",df.isnull().sum())

    for column in df.columns:
        if pd.api.types.is_numeric_dtype(df[column]) and column != target_name:
            df[column] = df[column].fillna(df[column].mean())
            df[column] = df[column].astype(float)
        else:
            df[column] = df[column].fillna(df[column].mode()[0])
    
    print("\nTotal Null value counts after imputation: \n",df.isnull().sum())
    return df

    
def time_series_imputation(time_series_df: pd.DataFrame, target_name: str) -> pd.DataFrame:
    """
        The function works on imputation for the given time series dataframe. For columns with
        numeric values, the null values of that column are imputed/filled through the method of linear 
        interpolation. For columns with string values, the null values of that column are imputed/filled
        with the mode (most frequent) value of that column. 

        Note: If the first row contains a null value in the numerical column then the imputation will not 
        work. 

        @parameters:

            time_series_df : pd.DataFrame:
                The dataframe of the time series dataset with the target column included.
            
            target_name: str:
                The name of the target column
        
        @return:
        
            Optional[pd.DataFrame] 
                It will return the imputed dataframe if the function is successful.
    """
    df = time_series_df.copy()

    first_row = df.iloc[0]
    has_nan_and_is_numeric = first_row.isnull() & first_row.apply(lambda x: isinstance(x, (float, int)))
    if has_nan_and_is_numeric.any():
        raise Exception('The first row cannot have a numeric null value.')
    
    print("Total Null value counts before imputation: \n",df.isnull().sum())

    for column in df.columns:
        if pd.api.types.is_numeric_dtype(df[column]) and column != target_name:
            df[column] = df[column].interpolate(method='linear')
            df[column] = df[column].astype(float)
        else:
            df[column] = df[column].fillna(df[column].mode()[0])
        
    print("\nTotal Null value counts after imputation: \n",df.isnull().sum())
    return df


def handle_imbalanced_data(df: pd.DataFrame, target_variable: str, strategy = "smote", k_neighbors=2, untouched_columns: List[str]=[]) -> pd.DataFrame:
    """
        The function balances a dataframe defined through a given sampling strategy to be
        ran on a classfication model.

        Note: Choosing "smote" as your strategy would give you an oversampled encoded dataframe. 

        @paramters:

            df: pd.DataFrame
                The dataframe for the given dataset.
            
            target_variable: str
                The name of the target column in the dataframe

            stratergy: str: default = SMOTE
                The name of the sampling stratergy:
                    - oversampling 
                    - undersampling 
                    - smote 
            
            k_neighbours: int: default = 2
                The number of neighbours for the Smote Stratergy

            untouched_columns : List[str]
                The list of column names that should not be encoded.
        
        @return:

            pd.DataFrame
                The balanced dataframe after following the sampling stratergy.
    
    """

    if not check_target_variable(df, target_variable):
        raise Exception(f"The target variable '{target_variable}' is missing in the DataFrame.")

    if strategy == 'oversampling':
        sampler = RandomOverSampler()
    elif strategy == 'undersampling':
        sampler = RandomUnderSampler()
    elif strategy == 'smote':
        sampler = SMOTE(k_neighbors=k_neighbors)
    else:
        raise ValueError("Invalid strategy. Choose from 'oversampling', 'undersampling', or 'smote'.")
    
    print(f'The distribution of the target column prior to sampling: {df[target_variable].value_counts}')
    
    if strategy == "smote":

        features, target = separate_target_column(df, target_variable)

        features = encoding(features, untouched_columns)

        columns_to_drop = [col for col in untouched_columns if col in features.columns]
        features_new = features.drop(columns=columns_to_drop)

        post_df = combine(features_new,target)

        X_res, y_res= separate_target_column(post_df,target_variable)

        X_res, y_res = sampler.fit_resample(X_res, y_res)

        df_balanced = pd.concat([X_res, y_res], axis=1)

        df_balanced = pd.concat([df_balanced, df[untouched_columns].reset_index(drop=True)], axis=1)

        for column in df_balanced.columns:
            if column in untouched_columns:
                df_balanced[column] = df_balanced[column].fillna(df_balanced[column].mode()[0])

    else:

        X_res, y_res= separate_target_column(df,target_variable)

        X_res, y_res = sampler.fit_resample(X_res, y_res)

        df_balanced = pd.concat([X_res, y_res], axis=1)

    print(f'The distribution of the target column after sampling: {df_balanced[target_variable].value_counts}')
    
    return df_balanced

def remove_outliers(df: pd.DataFrame) -> pd.DataFrame:
    """
        The function detects outliers outside the range of 2.575 (99% CI) and removes the corresponding
        observations/rows if the outlier is present in more than 10% of the column. 

        @paramters:

            df: pd.DataFrame
                The dataframe for the given dataset.
        
        @return:

            pd.DataFrame
                The balanced dataframe after following the sampling stratergy.

    """
    numerical_features = df.select_dtypes(include=['float64', 'int64']).columns
    df_original = df.copy()

    outlier_counts = pd.Series(0, index=df_original.index)
    threshold = 2.575 # Indicating a 99% CI
    col_count = len(numerical_features)

    for col in numerical_features:
        z_scores = np.abs(stats.zscore(df_original[col]))
        outliers = z_scores > threshold
        outlier_counts += outliers

    # Remove rows with a significant number of outliers.
    df_pre_processed = df_original[outlier_counts < col_count*0.1]

    print(f"\nOriginal data count: {len(df_original)}")
    print(f"After outlier removal data count: {len(df_pre_processed)}")

    return df_pre_processed

def separate_target_column(df: pd.DataFrame, target_variable: str) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """

        This function separates the given target column from the data frame.    

        @paramters:

            df: pd.DataFrame
                The dataframe for the given dataset that contains the target column as well.
            
            target_variable: str
                The name of the target column in the dataframe.
        
        @return:

            Tuple(pd.DataFrame, pd.DataFrame):
                Tuple[0] : The dataframe with all columns except the target column
                Tuple[1] : The target column dataframe.


    """
    try:
        if not isinstance(df, pd.DataFrame):
            raise TypeError
        
        if not check_target_variable(df, target_variable):
            raise ValueError

        target = df[[target_variable]]
        remaining_df = df.drop(columns=[target_variable])
        return remaining_df, target
    
    except TypeError:
        raise Exception('Data type error.Please provide the right type of dataframe.')
    
    except ValueError:
        raise Exception('Target Column does not Exist. Please provide the right one.')


def encoding(features: pd.DataFrame, untouched_columns: List[str]=[]) -> pd.DataFrame:
    """
        The function encodes the object type columns in the given data frame. If the number of
        attributes in a column exceeds more than 3, then the function performs Label Encoding. If it does not, 
        then it performs One Hot Encoding. 

        @parameters:
            features : str
                The data frame consisting of all the features (excluding the target column).
            
            untouched_columns : str
                The list consists of all the 
        
        @return:
            pd.DataFrame:
                The dataframe consisting of all the encoded features. 
    """
    print("Before encoding:", features.columns)
    for column in features.columns:
        if features[column].dtype == "object" and column not in untouched_columns:
            if features[column].nunique() > 3:
                encoder = LabelEncoder()
                features[column] = encoder.fit_transform(features[column])
            else:
                encoder = OneHotEncoder(sparse_output=False)
                encoded = encoder.fit_transform(features[[column]])
                encoded_features = pd.DataFrame(encoded, columns=encoder.get_feature_names_out([column]), index=features.index)
                features.drop(columns=[column], inplace=True)
                features[encoded_features.columns] = encoded_features
    
    print("After encoding:", features.columns)

    return features


def feature_scaling(features: pd.DataFrame, unscale_columns: List[str]) -> pd.DataFrame:
    """
        The function scales continuous-numerical values of the dataset. If the values in the column have a normal
        distribution then we do StandardScaling. If it does not, then we do MinMaxScaling
        Note : If the unique value count of the column is greater than 3, only then will the function apply 
        scaling to that column.

        @parameters:

            features:
                The dataframe containing the features to be scaled.
            
            unscale_columns:
                The list of column names that represent the columns that should not be scaled.


        @return:

            pd.DataFrame:   
                The dataframe containing the scaled features.

    
    """
    df = features.copy()
    
    numerical_columns = df.select_dtypes(include=['float64', 'int64']).columns
    non_numerical_columns = df.select_dtypes(exclude=['float64', 'int64']).columns

    def is_normal(column, alpha=0.05, max_sample_size=5000):
        column = column.dropna()

        if len(column) > max_sample_size:
            column = resample(column, n_samples=max_sample_size, random_state=42)

        anderson_result = stats.anderson(column, dist='norm')
        anderson_normal = anderson_result.statistic < anderson_result.critical_values[2]  

        skewness = stats.skew(column)
        kurtosis = stats.kurtosis(column)

        skew_normal = abs(skewness) < 0.5
        kurtosis_normal = abs(kurtosis) < 0.5

        # Return True if all conditions are met
        return anderson_normal and skew_normal and kurtosis_normal

    
    Minmaxscaler_algorithms = []
    Standardscaler_algorithms = []

    
    for col in numerical_columns:
        if col not in unscale_columns:

            unique_count = df[col].nunique()

            if is_normal(df[col]):
                
                if unique_count > 3: 
                    scaler = StandardScaler()
                    df[col] = scaler.fit_transform(df[[col]])
                    Standardscaler_algorithms.append(col)

            else:

                if unique_count > 3: 
                    scaler = MinMaxScaler()
                    df[col] = scaler.fit_transform(df[[col]])
                    Minmaxscaler_algorithms.append(col)
            
                        
    print("\nMinmax scaled columns: \n", Minmaxscaler_algorithms)
    print("\nStandardized columns: \n", Standardscaler_algorithms)
    
    df_scaled = pd.concat([df[numerical_columns], df[non_numerical_columns]], axis=1)
    
    df_scaled = df_scaled.reset_index(drop=True)

    return df_scaled

def combine (features: pd.DataFrame, target: pd.DataFrame) -> pd.DataFrame:
    """
        The functions combines the features dataframe and target dataframe into one dataframe.

        @paramters:
            features: pd.DataFrame
                The dataframe that consists of all the features
        
            target: pd.DataFrame
                The dataframe that consists of all the targets.
        
        @return:
            
            pd.DataFrame:
                The combined dataframe that consists of both the feature dataframe and the target dataframe.
    
    """
    try:
        if not isinstance(features, pd.DataFrame) and not isinstance(target, pd.DataFrame):
            raise TypeError
        combined_df = pd.concat([features, target], axis=1)
        return combined_df
    
    except TypeError:
        raise Exception('Data type error.Please provide the right type of dataframe.')


def preprocess_pipeline(file_path: str, target_column: str, dropped_columns: List[str]=[], untouched_columns: List[str]=[], type_dataset: int=0, sampling: int=0, classfication: int=1, strategy_sample="smote"):
    """
        This function is a stand-alone pipeline used for the free website template where you can 
        interact with a GUI.
        
        @parameters:

            file_path: str 
                The path to the dataset.

            target_column: str 
                The name of the target variable.

            dropped_columns: List[str]
                The list of column names the client wants dropped from the dataset.

            untouched_columns: List[str]
                The list of columns that should be not scaled/encoded.

            type_dataset: int
                The type of dataset provided:
                    - 0 -> Cross Sectional
                    - 1 -> Time Series
            
            sampling: int 
                Indicates whether the client wants to perform sampling or not
                    - 0 -> No sampling
                    - 1 -> Sampling 
                    
            classification: int 
                The type of model that the dataset will be trained on 
                Note: The pipeline does not perform sampling for regression models.
                    - 0 -> regression
                    - 1 -> classification

            stratergy_sample: str: default = SMOTE
                The name of the sampling stratergy:
                    - oversampling 
                    - undersampling 
                    - smote 

        @return:

            pd.DataFrame:
                The proccessed dataframe.

    """
    print("\nInitial Check")
    df = initial_check_dt(file_path, target_column, dropped_columns)
    print("\n-----------------------------Initial check done-----------------------------------------------\n")
    
    print("\nImputation")
    if not type_dataset:
        df = cross_sectional_imputation(df, target_column)
        print("\n-----------------------------Cross Sectional Imputation Done-----------------------------------------------\n")
    elif type_dataset == 1:
        df = time_series_imputation(df, target_column)
        print("\n-----------------------------Time Series Imputation Done-----------------------------------------------\n")

    if sampling:
        if classfication:
            if df[target_column].nunique() > 2:
                print("\n-----------------------------Balancing Operations not suported for Multi-class Classification-----------------------------------------------\n")
            else:
                if strategy_sample == "smote":            
                    print("\nHandling Imbalanced Data")
                    df = handle_imbalanced_data(df, target_column, strategy_sample, untouched_columns=untouched_columns)
                    print("\n-----------------------------Balanced and Encoded the data-----------------------------------------------\n")
                else:
                    print("\nHandling Imbalanced Data")
                    df = handle_imbalanced_data(df, target_column, strategy_sample)
                    print("\n-----------------------------Balanced the data-----------------------------------------------\n")
        else:
           print("\n-----------------------------Balancing Operations not suported for Regression Models-----------------------------------------------")
           print("\n-----------------------------Did not Balance the data-----------------------------------------------\n")

    print("\nRemoving Outliers")
    df = remove_outliers(df)
    print("\n-----------------------------Removed Outliers-----------------------------------------------\n")
    
    print("\nSeparating Input and Output")
    remaining_df, target = separate_target_column(df, target_column)
    print("\n-----------------------------Separated the input and output-----------------------------------------------\n")
    
    
    if strategy_sample != "smote" or df[target_column].nunique() > 2:
        print("\nEncoding Data")
        remaining_df = encoding(remaining_df, untouched_columns)
        print("\n-----------------------------Encoded the data-----------------------------------------------\n")
    
    print("\nFeature Scaling")
    preprocessed_df = feature_scaling(remaining_df, untouched_columns)
    print("\n-----------------------------Feature Scaling Done-----------------------------------------------\n")
    
    print("\nCombining Feature and Target")
    combined_df = combine(preprocessed_df,target)
    print("\n-----------------------------Combining Features Done-----------------------------------------------\n")
    
    return combined_df



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    description = f.read()

setup(
    name='weedout', 
    version='1.2', 
    packages=find_packages(), 
    install_requires=[
        'pandas',
        'numpy',
        'seaborn',
        'matplotlib',
        'scikit-learn',
        'statsmodels',
        'imbalanced-learn',
        'scipy',
        'typing'
    ],
    long_description=description,
    long_description_content_type ="text/markdown",

)


import pandas as pd
from statsmodels.stats.outliers_influence import variance_inflation_factor


def filtered_correlation_matrix(df: pd.DataFrame):
    """
        This function prints the Variance Inflation Factor ( VIF = 1/(1-R^2) ) of each feature column. VIF 
        is a strong indicator of multi-collinearity in our data frame. 

        Note : The user is expected to remove some of the correlated features based on the VIF value.

        @parameter:

            df : pd.DataFrame
                The dataframe provided by user.
        
    """

    numerical_columns = df.select_dtypes(include=['float64', 'int64'])
    print("Before: ", numerical_columns.shape)
    
    vif = pd.DataFrame()
    vif['Feature'] = numerical_columns.columns
    vif["VIF"] = [variance_inflation_factor(numerical_columns.values, i) for i in range(numerical_columns.shape[1])]
    
    print("Initial VIF values: ")
    print(vif.sort_values(by="VIF", ascending=False))


def display(file_path: str ,df: pd.DataFrame) -> None:
    """
        The function prints the info of the original dataset in the given file path in comparison to
        the preprocessed dataframe.

        @paramters:
            file_path: str 

                The path to the original file.

            df: pd.DataFrame 

                The preproccsed dataframe.

    """
    original_data=pd.read_csv(file_path)
    processed_data=df
    print("\nOriginal Data\n")
    original_data.info()
    print("\nProcessed Data\n")
    processed_data.info()